/** 
 *   Site testing for SKA candidate.
 *
 *   (c) Copyright 2004 Federico Bareilles <fede@iar.unlp.edu.ar>,
 *   Instituto Argentino de Radio Astronomia (IAR).
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *     
 *   The author does NOT admit liability nor provide warranty for any
 *   of this software. This material is provided "AS-IS" in the hope
 *   that it may be useful for others.
 *
 **/
#define _GNU_SOURCE

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>
#include <sys/file.h>
#include <string.h>

#include "my_stdio.h"
#include "my_math.h"
#include "rotator.h"
#include "observ_table.h"

#define PK_LEN   5
/* command string definition: */
#define PK_STX   0
#define PK_CMD   1 
#define PK_DATA1 2 
#define PK_DATA0 3
#define PK_CHS   4 

#define STX_DATA    0xff  /* Start of Transmision */
#define DC_DATA     0x00  /* Don't care */

#define SET_RESET   0x00
#define SET_AZ      0x03
#define SET_POL     0x01

#define ROT_ERR     0x02

#define GET_AZ      0x05
#define GET_POL     0x04
#define GET_ERR     ROT_ERR

#define POL_HORIZONTAL  0
#define POL_VERTICAL    1
#define POL_INDEF       2

#define TO_ROTATOR_L(v) (v & 0xff)
#define TO_ROTATOR_H(v) ((v >> 8) & 0x7f)
#define FROM_ROTATOR(pk) ((((pk[PK_DATA1] & 0xff) << 8) | (pk[PK_DATA0] & 0xff)) & 0x7fff)

#define ROT_AZ_CONST (1000.0/54.0)
#define ROT_AZ_OFFSET (10000.0)
#define DEG2ROT(a) ((int)my_round((double) a * ROT_AZ_CONST + ROT_AZ_OFFSET))
#define ROT2DEG(a) ((int)my_round(((double) a - ROT_AZ_OFFSET) / ROT_AZ_CONST))


static int rt_check_packet(char *pk);


int rotator_setup(char *name)
{
	int fd;
	struct termios tio;

#ifdef DEMO_MODE 
	return 2;
#endif
	/* Open Serial Port : 8 bits data, 1 bit stop, No parity, 1200
           baud */
	if ((fd = open(name, O_RDWR | O_NOCTTY)) < 0 ){
		SYSLOG(LOG_ERR, "Cannot open serial port (%s).", name);
		return -1;
	}

	memset(&tio, 0, sizeof(tio));
	tio.c_cflag &= ~CRTSCTS ;
	/* Zero out the baud rate portion of c_cflag */
	tio.c_cflag &= ~CBAUD;
	/* Set baud rate to 1200 */
	tio.c_cflag |= B1200;
	tio.c_cflag &= ~(CSIZE|PARENB);
	tio.c_cflag |= CS8 ;
	/* Ignore modem control lines */
	tio.c_cflag |= CLOCAL;
	/* Enable receiver */
	tio.c_cflag |= CREAD;

	/* Turn of Canonical Processing and echo */
	tio.c_lflag &= ~ICANON;
	tio.c_lflag &= ~ECHO;
	/* Set MIN to one and TIME to zero */
	tio.c_cc[VMIN]  = 1;
	tio.c_cc[VTIME] = 0;

	/* Ignore BREAK condition on input */
	tio.c_iflag |= IGNBRK;
	/* ignore framing errors and parity errors */
	tio.c_iflag |=IGNPAR;
	/* Disable NL to CR mapping */
	tio.c_iflag &= ~ICRNL;
	/* Disable XON/XOFF flow control on output and input */
	tio.c_iflag &= ~(IXON|IXOFF);

	/* now clean the modem line and activate the settings for modem */
	tcflush(fd, TCOFLUSH|TCIFLUSH);
	tcsetattr(fd, TCSANOW, &tio);
	
	return fd;
}

int rotator_close(int fd)
{
#ifndef DEMO_MODE
	rotator_set_azimuth(fd, ROTATOR_AZ_PARK);
	close(fd);
#endif
	return 0;
}


static int rt_log_bad_packet(void)
{
	SYSLOG(LOG_ERR, "Bad packet from rotator.");

	return -1;
}


static ssize_t soft_read_time_out(struct timeval tv, int fd, void *buf, 
				  size_t count)
{
	fd_set rfds;
	int fd_max = fd + 1;
	int retval, l = 0;
	struct timeval tvl;

	while ( l < count) {
	  memcpy(&tvl, &tv, sizeof(struct timeval));
		FD_ZERO(&rfds);
		FD_SET(fd, &rfds);
		retval = select(fd_max, &rfds, NULL, NULL, &tvl);
		if (retval) {
			if( FD_ISSET(fd, &rfds) ) {
			    retval = read(fd, buf+l, count - l);
				l += retval;
			}
		}else{
			return ROT_ERR_TIME_OUT;
		}
	}

#ifdef DEBUG_MODE
	if (verbose_flag & PINFO) {
		int __i;
		PVERB(PINFO, "# rotator.c read <- ");
		for(__i=0;__i<l;__i++)
			PVERB(PINFO, "0x%02x ", 
			      *(((u_char *)buf)+__i) );
		PVERB(PINFO, "\n");
	}
#endif	
	
	return l;
}

#ifndef DEMO_MODE
static int discard_noise(int fd)
{
	struct timeval tv;
	char noise;
	int i;
	char pk[PK_LEN];

	tv.tv_sec = 0;
	tv.tv_usec = 1;
	i = 0;
	while ( soft_read_time_out(tv, fd, &noise, 1) > 0) {
		SYSLOG(LOG_WARNING,
		       "rotator: noise 0x%02x", (u_char) noise);
		if ( i && i < PK_LEN) {
			pk[i] = noise;
			i++;
		}
		if ( i == PK_LEN ) {
			if( rt_check_packet(pk) == 0) {
				/* is real packet, not noise */
				/* TODO */
				SYSLOG(LOG_WARNING,
				       "rotator: real packer arrived (?).");
				i = 0;
			}
		}
		if ( noise == (char) STX_DATA ) {
			pk[0] = STX_DATA;
			i = 1;
		}
	}
	
	return 0;
}
#endif

static ssize_t soft_read(int fd, void *buf, size_t count)
{
	struct timeval tv;

	tv.tv_sec = ROTATOR_ECHO_TIME_OUT;
	tv.tv_usec = 0;

	return soft_read_time_out(tv, fd, buf, count);
}


static ssize_t soft_write(int fd, const void *buf, size_t count)
{
#ifndef DEMO_MODE
        discard_noise(fd);
#endif
#ifdef DEBUG_MODE
	if (verbose_flag & PINFO) {
	       int __i;

		PVERB(PINFO, "# rotator.c send -> ");
		for(__i=0;__i<count;__i++)
			PVERB(PINFO, "0x%02x ", *(((u_char *)buf)+__i) );
		PVERB(PINFO, "\n");
	}
#endif
#ifdef DEMO_MODE
	return count;
#endif
	return write(fd, buf, count);

}

	
static int rt_check_packet(char *pk)
{
	int i;
	char chs = pk[PK_STX];
	
	for (i=1; i < PK_LEN - 1; i++) chs ^= pk[i];
	if ((pk[PK_STX] == (char) STX_DATA) && (chs == pk[PK_CHS])) {
		return 0;
	}
	rt_log_bad_packet();

	return ROT_ERR_INVALID_PK;
}


static int rt_wait_confirmation(int fd, char *pk)
{
	struct timeval tv;
	char answer[PK_LEN];
	int res;

	tv.tv_sec = ROTATOR_TIME_OUT;
	tv.tv_usec = 0;	
	if( soft_read_time_out(tv, fd, answer, 5) < 0 ) {
		SYSLOG(LOG_ERR, "rotator: wait confirmation time out, pkcmd: 0x%x.",
		       pk[PK_CMD]);
	}else{
		if((res = strncmp( answer, pk, PK_LEN)) != 0) {
			if (rt_check_packet( answer ) == 0) {
				/* valid packet */
				switch (answer[PK_CMD]) {
				case ROT_ERR:
					res = ROT_ERR_CONDITION;
					SYSLOG(LOG_ERR,
					       "rotator: error condition.");
					/* strncpy(pk, answer, PK_LEN); */
					break;
				default:
					SYSLOG(LOG_ERR,
					       "rotator: packet out of sequence, pkcmd: 0x%02x", 
					       answer[PK_CMD]);
#ifdef DEBUG_MODE
					SYSLOG(LOG_INFO, 
					       "rotator.c read <- 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x", 
					       answer[PK_STX], answer[PK_CMD],
					       answer[PK_DATA1], 
					       answer[PK_DATA0], 
					       answer[PK_CHS]);
#endif	



					res = -1;
				}
			}else{
				res = ROT_ERR_INVALID_PK;
			}
		}
		return res;
	}
	
	return ROT_ERR_TIME_OUT;
}
	

static int rt_make_packet(char *pk, char cmd, char data1, char data0)
{
	int i;
	
	pk[PK_STX] = STX_DATA;
	pk[PK_CMD] = cmd;
	pk[PK_DATA1] = data1;
	pk[PK_DATA0] = data0;
	pk[PK_CHS] = pk[PK_STX];
	for (i=1; i< PK_LEN - 1; i++) 
		pk[PK_CHS] ^= pk[i];

	return 0;
}


static int rt_rcv_packet(int fd, char *pk)
{
	int res;

	res = soft_read(fd, pk, PK_LEN);
	if (res == ROT_ERR_TIME_OUT) {
		SYSLOG(LOG_ERR, "rotator: read packet time out.");
	}

	return res;
}


static int rt_send_packet(int fd, char *pk)
{
	char answer[PK_LEN];
	int res = -1;
	int i;

	for (i=0; i < ROTATOR_RETRY_SEND && res != 0; i++) {
		soft_write(fd, pk, PK_LEN);
		res = rt_rcv_packet(fd, answer);
		if (res == ROT_ERR_TIME_OUT) {
			break;
		}
		if((res = strncmp( answer, pk, PK_LEN)) != 0) {
			rt_log_bad_packet();
		}
	}

	return res;
}


/******************************************************************************
 * Rotator commands:
 **/


int rotator_reset(int fd)
{
	char pk[PK_LEN];
	int res;

#ifdef BAD_ROTATOR
	int az;

	if ((res = rotator_get_azimuth(fd, &az)) != 0) {
		return res;
	}
	if (az <=0) {
		if ((res = rotator_set_azimuth(fd, 10)) != 0) {
			return res;
		}
	}
#endif
	rt_make_packet(pk, SET_RESET, DC_DATA, DC_DATA);
	if ((res = rt_send_packet(fd, pk)) != 0) 
		return res;
	
	return rt_wait_confirmation(fd, pk);
}


int rotator_set_azimuth(int fd, int az)
{
	char pk[PK_LEN];
	int res;

	az = DEG2ROT(az);
#ifdef DEBUG_MODE
	PVERB(PINFO,"# az %5d\n", az);
#endif
	rt_make_packet(pk, SET_AZ, TO_ROTATOR_H(az), TO_ROTATOR_L(az));
	if ((res = rt_send_packet(fd, pk)) != 0) 
		return res;

	return rt_wait_confirmation(fd, pk);
}


int rotator_set_polarization(int fd, char pol)
{
	char pk[PK_LEN];
	int res;
	char poln;
	
	if (pol == HORIZONTAL) poln = POL_HORIZONTAL;
	else if (pol == VERTICAL) poln = POL_VERTICAL;
	else return -1;
	
	rt_make_packet(pk, SET_POL, DC_DATA, TO_ROTATOR_L(poln));
	if ((res = rt_send_packet(fd, pk)) != 0) 
		return res;

	return rt_wait_confirmation(fd, pk);
}


int rotator_get_azimuth(int fd, int *az)
{
	char pk[PK_LEN];
	int res;
        int v;
	
	rt_make_packet(pk, GET_AZ, DC_DATA, DC_DATA);
	if ((res = rt_send_packet(fd, pk)) != 0) {
		return res;
	}
	if ( rt_rcv_packet(fd, pk) < 0 ){
		return ROT_ERR_TIME_OUT;
	}
	if ((pk[PK_CMD] != (char) GET_AZ) || rt_check_packet(pk) != 0) {
		return ROT_ERR_INVALID_PK;
	}
	v = FROM_ROTATOR(pk);
#ifdef DEBUG_MODE
	PVERB(PINFO,"# az = %5d counts, %5.2f deg\n",v,
			(((double) (v - ROT_AZ_OFFSET)) / ROT_AZ_CONST)
			);
#endif
	*az = ROT2DEG(v);

	return 0;
}


int rotator_get_polarization(int fd, char *pol)
{
	char pk[PK_LEN];
	int res;

	rt_make_packet(pk, GET_POL, DC_DATA, DC_DATA);
	if ((res = rt_send_packet(fd, pk)) != 0) 
		return res;
	if ( rt_rcv_packet(fd, pk) < 0 ){
		return ROT_ERR_TIME_OUT;
	}
	if ((pk[PK_CMD] != (char) GET_POL) || rt_check_packet(pk) != 0) {
		return ROT_ERR_INVALID_PK;
	}
	if(pk[PK_DATA0] == POL_HORIZONTAL)
		*pol = HORIZONTAL;
	else if(pk[PK_DATA0] == POL_VERTICAL)
		*pol = VERTICAL;
	else
		*pol = INDEF;

	return 0;
}


